;;; **********************************************************************
;;; ** THOR Basic++                                                     **
;;; ** A free Basic for the Atari 8 Bit series               		**
;;; ** (c) 2015 THOR Software, Thomas Richter                           **
;;; ** This is based on Atari Basic, as developed by			**
;;; ** Paul Laughton and Kathleen O'Brien, Shepardson Microsystems	**
;;; ** as found in: The Atari BASIC SOURCE BOOK	by Bill Wilkinson	**
;;; ** Compute! Publications (1983)					**
;;; ** $Id: basicmain.asm,v 1.15 2015/09/15 14:00:28 thor Exp $		**
;;; **                                                                  **
;;; ** In this module:   Main program					**
;;; **********************************************************************

	.include "basicmacros.i"
	.include "basicmemory.i"
	.include "basicstack.i"
	.include "basicexecution.i"
	.include "basicparser.i"
	.include "basicio.i"
	.include "basicexpression.i"
	.include "basicstatements.i"
	.include "basicmain.i"
	.include "syntax.i"
	.include "constants.i"
	.include "mathpack.i"
	.include "reset.i"
	.include "editor.i"

	.segment "BasMain"

;;; *** BasicColdStart
;;; *** The basic enters here on a reset
	.global BasicColdStart
.proc	BasicColdStart
	lda LoadFlag		;program is corrupt? Try, try again.
	bne cold1
	inc GenerationCounter	;restarting the system, absolute addresses may get wrong
	lda WarmStartFlag
	bne WarmerStart
cold1:				;ok, completely from start, NEW out the system.
	pha			;keep the loadflag
	jsr InitializeMemory	;also resets the generation counter
	pla
	bne WarmerStart		;if the load flag was set, do not try again
	lda #>TitleMsg
	ldy #<TitleMsg
	jsr PrintStringAY
	jsr RunAutorun
	;; runs into the following
.endproc
;;; *** WarmerStart (SNX1 in Atari Basic)
;;; *** Closes all channels, resets stdin/stdout
;;; *** but does not RESTORE
	.global WarmerStart
.proc	WarmerStart
	jsr CloseAll
	jsr ResetEnterListIOCB	;select IOCB #0
	;; runs into the following
.endproc
;;; *** ExpectInput (SNX3 in AtariBasic)
;;; *** Print the READY prompt and await input to be parsed, then parse
	.global ExpectInput
.proc	ExpectInput
	jsr PrintCR
	lda #>ReadyMsg
	ldy #<ReadyMsg
	jsr PrintStringAY	;Print the famous READY string
	;; runs into the following
.endproc
;;; *** Syntax (SYNTAX in Atari Basic)
;;; *** This is the main loop of the Atari Basic editor mode, i.e.
;;; *** the place where statements are received and executed in direct mode,
;;; *** or where lines are read, tokenized and inserted into the statement
;;; *** table.
	.global Syntax
.proc	Syntax
	ldx #$ff		;reset stack
	txs
	lda LoadFlag		;interrupted and jumped into here while loading was in progress?
	bne BasicColdStart
	jsr LoadOutbuffVector	;load OutBuf into the inbuf vector
	jsr GetLine		;read a line from the editor
	jsr TestForBreak	;BREAK key pressed?
	bne Syntax		;if so, abort parsing this line
	;; prepare now parsing the input buffer
	lda #0
	sta cix			;offset in the input buffer
	sta MaxValidcix		;clear this as well
	sta cox			;reset output buffer index
	sta DirectModeFlag	;execut directly or insert a line?
	sta LastOperatorcix	;no operator parsed yet
	sta ParsedOperator	;clear the token backup
	sta AdditionalVariables
	;; make a copy of the endptr
	lda VarNameEndPtr
	sta PreviousVarNameEndPtr
	lda VarNameEndPtr+1
	sta PreviousVarNameEndPtr+1
	;; parsing starts now
	jsr SkipBlanksVector	;skip over blanks in the input buffer (advancing cix)
	jsr ParseLineNumber	;try to parse off the line number
	jsr InsertToken		;reserve space for the line length
	lda TargetLineNumber+1	;was this a line number >32768 and hence are we in direct mode?
	bpl isedit
	sta DirectModeFlag	;yes, is a direct command to be executed immediately
isedit:
	jsr SkipBlanksVector	;skip over the following blanks
	ldy cix
	sty StatementStart 	;keep me, here we start looking for statements
	lda (inbuff),y		;get next character
	cmp #$9b		;is it EOL?
	bne ParseNextStatement	;got an empty line?
	bit DirectModeFlag	;are we in direct mode?
	bmi Syntax		;if so, just an empty line, do nothing
	jsr DeleteLine		;otherwise, delete the line
	jmp Syntax		;continue parsing
.endproc
;;; *** ParseNextStatement (XIF in Atari Basic)
;;; *** This is the main loop that parses off statement after
;;; *** statement in a line
	.global ParseNextStatement
.proc	ParseNextStatement
tokenloop:
	lda cox
	sta StatementLen 	;keep offset of the length byte
	jsr InsertToken		;reserve room for the statement length
	jsr SkipBlanksVector	;more blanks to avoid
	
	lda #>StatementTable
	ldy #<StatementTable
	ldx #2			;two extra bytes per token
	jsr SearchTable
	;; if this does not find the statement, it will use the token of the ERROR statement...
	stx cix			;update cix, points now behind the statement (or at statement, if not found)
	lda ScanOffset		;this is now the statement offset
	jsr InsertToken		;add to output buffer
	jsr SkipBlanksVector
	jsr ParseArguments	;parse now the arguments of the statement
	bcc syntaxok
	;; here: invalid syntax, mark with ERROR
	;; check how far we got with the syntax parsing
	jsr InsertError
syntaxok:
	;; now fixup the statement end offset
	lda cox			;end position of the current statement
	ldy StatementLen
	sta (TokenBuffer),y	;adjust the end offset of the statement.
	;; check now whether we reached the end of the line (and not only the statement)
	ldy cix
	dey
	lda (inbuff),y
	cmp #$9b		;found EOL here?
	bne tokenloop
	;; runs into the following
.endproc
;;; *** FinalizeLine
;;; *** Insert the line into the statement table, potentially execute it.
	.global FinalizeLine
.proc	FinalizeLine
	;; ok, here the line is complete.
	ldy #2			;this is where we need to set the line size
	lda cox
	sta (TokenBuffer),y	;insert here
	;; insert into the statement table
	jsr InsertLine

	;; check for synax errors or direct execution
	bit DirectModeFlag
	bvc noerror
	;; here we have found a parsing error
	lda AdditionalVariables	;remove variables allocated on the go?
	asl a
	asl a
	asl a
	tay			;8 bytes per variable
	ldx #StatementPtr	;move statement table down
	jsr FreeMemLow		;and release the new values again
	sec
	lda VarNameEndPtr
	sbc PreviousVarNameEndPtr
	tay
	lda VarNameEndPtr+1
	sbc PreviousVarNameEndPtr+1
	ldx #VarNameEndPtr
	jsr FreeMem		;and release also the variable names that came here due to error

	bit DirectModeFlag	;are we in direct mode?
	bpl notdirect
	;; here direct mode
	jsr ListDirectLine	;list the direct line
	bcs gosyntax		;return to the main loop
notdirect:
	jsr ListLine		;list the faulty line
	bcs gosyntax
noerror:
	bpl gosyntax		;if not direct, just continue parsing
	;; here: A direct command, execute now
	jmp ExecuteProgram
gosyntax:	
	jmp Syntax		;repeat the main loop
.endproc
;;; *** InsertLine (:SYNOK in Atari Basic, but not a subroutine)
;;; *** Insert the currently parsed line into the statement table
;;; *** move the statement table as necessary
.proc 	InsertLine
	;; now check whether we possibly need to replace the line in the statement table
	jsr FindTargetLine
	lda #0			;size of the old line if none found
	bcs insertnew		;branch if not present
	;; here: replace line
	jsr GetLineLength	;length of the old line
insertnew:			;or zero if there was no previous line before
	sec
	sbc cox			;do we need to reserve more or release memory?
	beq replace		;same storage requirment->copy over
	bcs release		;old line size is larger->release memory
	eor #$ff
	tay
	iny			;two's complement: number of bytes to reserve
	ldx #CurrentLinePtr	;move in here
	jsr AllocMemLow		;reserve memory
	lda AllocMemPtr		;get the address of the newly available memory
	sta CurrentLinePtr
	lda AllocMemPtr+1
	sta CurrentLinePtr+1
	bne replace		;and copy the line parsed in just into there
release:			;here: new line is shorter than old line
	pha			;keep the differential size
	jsr GetNextLine		;move the next line down
	pla
	tay			;byte size
	ldx #CurrentLinePtr
	jsr FreeMemLow		;and release the memory
replace:			;here: memory has been moved, now copy data over
	ldy cox
cploop:
	dey
	lda (TokenBuffer),y
	sta (CurrentLinePtr),y
	tya
	bne cploop		;until the last byte copied
	rts
.endproc
;;; *** DeleteLine (:SDEL in Atari Basic, but not a subroutine)
;;; *** Delete the line given by TargetLineNumber
.proc	DeleteLine
	jsr FindTargetLine	;go search for the line
	bcs nosuchline		;if not found, do not bother
	jsr GetLineLength	;extract the length of this line
	pha			;store
	jsr GetNextLine		;advance to the end
	pla
	tay			;move line length to Y again
	ldx #CurrentLinePtr	;move this line out of memory
	jsr FreeMemLow
nosuchline:
	rts
.endproc


;;; *** ListLine (LLINE in Atari Basic)
;;; *** list an entire line pointed to by CurrentLinePtr
;;; *** including the line number
	.global ListLine
.proc	ListLine
	jsr PrintLineNumber
	;; runs into the following
.endproc
;;; *** ListDirectLine (LPLINE in Atari Basic)
;;; *** Print the line excluding the line number
;;; *** Returns with C set!
.proc	ListDirectLine
	jsr GetLineLength
	sta LineLength
	iny
listloop:
	lda (CurrentLinePtr),y
	sta StatementEnd	;get offset of next statement
	iny
	sty StatementStart 	;start of current statement
	jsr ListStatement	 ;print the statement at this offset
	ldy StatementEnd 	;get next offset
	cpy LineLength
	bcc listloop		;continue until the line is done
	rts
.endproc
;;; *** ListStatement (LSTMT in Atari Basic)
;;; *** List the statement at CurrentLinePtr, at offset StatementStart
.proc	ListStatement
	jsr GetListToken
	cmp #ImplicitLetToken	;variable directly?
	beq printargs
	jsr PrintStatement	;print the statement
	jsr GetListToken
	cmp #ErrorToken		;the token of ERROR -
	beq outputdata		;output the following data directly
	cmp #INPUTToken		;all regular commands
	bcs printargs
	;; here REM or DATA: This prints the data directly until the end of the data
outputdata:			;or ERROR
	jsr GetNextListToken
	jsr PrintChar		;print the final EOL
	bpl outputdata		;aborts by GetNextListToken
	;; here a statement that takes tokenized arguments (the regular case)
printargs:
	jsr GetNextListToken
	bpl isoperand
	;; here: is a variable
	and #$7f
	sta ScanOffset		;the variable number
	ldx #0			;zero extra bytes in the table
	lda VarNamePtr+1	;variable names
	ldy VarNamePtr
	jsr SearchForToken	;look for the variable name
	jsr PrintString		;and print the variable name
	cmp #'('+$80		;an array?
	bne printargs		;if not, continue printing the expression
	;; here: an array variable that has to be printed
	;; next token is a ( anyhow and does not need printing
	jsr GetNextListToken
	bcc printargs		;jumps always
isoperand:
	cmp #StringToken 	;a constant string in quotes?
	beq printstringconst
	bcs printoperand
	;; here: a constant number
	jsr GetNextBCD		;read a BCD number from the statement
	sty StatementStart 	;restore the token pointer
	jsr BCDToAsciiVector	;convert number to ASCII
	jsr LoadinbuffToPrintPtr
nonblank:			;used below. print operand directly
	jsr PrintString		;and print the number
	clc
	bcc printargs		;continue with the tokens
	;; print a string constant
printstringconst:
	jsr GetNextListToken	;read the next token
	sta PrintOffset		;is string length
	lda #$22		;double quote
	jsr PrintChar
	lda PrintOffset
	beq prdone
prstring:	
	jsr GetNextListToken
	jsr PrintChar		;print the contents of the string
	dec PrintOffset		;until done
	bne prstring
prdone:
	lda #$22		;another double quote
	jsr PrintChar
	bpl printargs		;and continue
	;; print a tokenized operand
printoperand:
	sbc #CommaToken		;C=1. Compute the offset in the operand table
	sta ScanOffset
	ldx #0			;no additional entries
	lda #>OperatorTable
	ldy #<OperatorTable
	jsr SearchForToken	;find token name
	jsr GetListToken	;get again
	cmp #STRToken		;regular functions?
	bcs nonblank		;without blanks around them, please
	ldy #0
	lda (ScanPtr),y		;get the name of the beast
	and #$7f
	jsr IsAlpha		;can this be printed?
	bcs nonblank		;print without spaces around it (things like +,-,*,/...)
	jsr PrintBlankStringAndBlank ;otherwise, put blanks around it
	bpl printargs		;continue
	;; code never goes here, returns by falling off GetNextListToken
.endproc
;;; *** GetNextListToken (LGNT in Atari Basic)
;;; *** inrement the in-token offset, get the next token
.proc	GetNextListToken
	inc StatementStart
	;; runs into the following
.endproc
;;; *** GetListToken (LGCT in Atari Basic)
;;; *** Get the token at CurrentLinePtr + StatementStart
;;; *** return to the caller of the caller if the end of the statement is reached
.proc	GetListToken
	ldy StatementStart
	cpy StatementEnd 	; below end?
	bcs abort
	lda (CurrentLinePtr),y
	rts
abort:
	pla			;return to the caller
	pla
	rts
.endproc
;;; *** PrintStatement (AtariBasic LSTMC)
;;; *** Print the command token whose token index is in register A
	.global PrintStatement
.proc	PrintStatement
	sta ScanOffset		;store the token number
	ldx #2			;skip two bytes after each token
	lda #>StatementTable
	ldy #<StatementTable
	jsr SearchForToken	;find the token in a token table (LSCAN)
	;; returns the token address in PrintPtr
	jmp PrintStringAndBlank	;print the token, followed by a blank space
.endproc

